import logging
import re
import yaml

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .tag_images import tag_images
from .utils.misc import (
    run_entrypoint,
    get_internal_container_repo_name,
    log_step,
)

LOG = logging.getLogger("PubLogger")
logging.basicConfig()
LOG.setLevel(logging.INFO)


class OperatorPusher:
    """
    Add operator bundles to index images and push them to Quay.

    No validation is performed, push items are expected to be correct.
    """

    def __init__(self, push_items, target_settings):
        """
        Initialize.

        Args:
            push_items ([ContainerPushItem]):
                List of push items.
            target_name (str):
                target name
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.target_settings = target_settings

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")

    @staticmethod
    def _get_immutable_tag(push_item):
        """
        Return immutable tag from operator push item production tags.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Immutable tag
        """
        # NOTE: while this was copied from pub, '[0]' had to be added here to work properly
        if push_item.metadata["v_r"] in list(push_item.metadata["tags"].values())[0]:
            return push_item.metadata["v_r"]
        # if v_r tag is not in destination tags, return tag with the most numbers in it
        # This code is usually trigerred when push is not initiated from ET but manually
        # by user
        tags = []
        for tag in list(push_item.metadata["tags"].values())[0]:
            tags.append((len(re.split(r"\d+", tag)), tag))

        return sorted(tags)[-1][1]

    def public_bundle_ref(self, push_item):
        """
        Get public reference of a bundle image.

        It will be used by IIB to access the bundle image.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Customer-visible bundle reference.
        """
        repository = list(push_item.metadata["tags"].keys())[0]
        # tags are the same for each destination repo, so any combination should work
        return "{0}/{1}:{2}".format(
            self.target_settings["docker_settings"]["docker_reference_registry"][0],
            repository,
            self._get_immutable_tag(push_item),
        )

    def pyxis_get_ocp_versions(self, push_item):
        """
        Get a list of supported ocp versions from Pyxis.

        Args:
            push_item: (ContainerPushItem)
                Push item for which the OCP version range will be found out.

        Returns ([str]):
            Supported OCP versions as returned by Pyxis.
        """
        ocp_versions = push_item.metadata["com.redhat.openshift.versions"]
        LOG.info("Getting OCP versions of '{0}' from Pyxis.".format(ocp_versions))

        args = ["--pyxis-server", self.target_settings["pyxis_server"]]
        args += ["--pyxis-krb-principal", self.target_settings["iib_krb_principal"]]
        args += ["--organization", self.target_settings["iib_organization"]]
        args += ["--ocp-versions-range", ocp_versions]
        if "iib_krb_ktfile" in self.target_settings:
            args += ["--pyxis-krb-ktfile", self.target_settings["iib_krb_ktfile"]]
        env_vars = {}

        data = run_entrypoint(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"),
            "pubtools-pyxis-get-operator-indices",
            args,
            env_vars,
        )

        if not data:
            msg = "Pyxis has returned no OCP versions for '{0}' specified in build {1}.".format(
                ocp_versions, push_item.metadata["build"]["build_id"]
            )
            raise ValueError(msg)

        # Versions returned by Pyxis don't contain 'v' at the front (4.5 -> v4.5)
        return ["v{0}".format(item["ocp_version"]) for item in data]

    def generate_version_items_mapping(self):
        """
        Generate mapping of OCP version -> push_items.

        The mapping describes which operator bundles should be added to which index images.

        Returns ({str: [ContainerPushItem]})
            Mapping of OCP version -> Push items
        """
        version_items_mapping = {}
        ocp_versions_resolved = {}

        for item in self.push_items:
            ocp_versions = item.metadata["com.redhat.openshift.versions"]
            # we haven't yet encountered this pattern, contact Pyxis for resolution
            if ocp_versions not in ocp_versions_resolved:
                ocp_versions_resolved[ocp_versions] = self.pyxis_get_ocp_versions(item)

            for version in ocp_versions_resolved[ocp_versions]:
                version_items_mapping.setdefault(version, []).append(item)

        return version_items_mapping

    def get_deprecation_list(self, version):
        """
        Get bundles to be deprecated in the index image.

        Args:
            version: (str)
                version for which deprecation list will be fetched.

        Returns:
            list(str): list of bundles to be deprecated in the index image.
        """

        def _get_requests_session():
            session = requests.Session()
            retry = Retry(
                total=6,
                read=6,
                connect=6,
                backoff_factor=0.8,
                status_forcelist=(500, 502, 503, 504),
            )
            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)

            return session

        deprecation_list = []
        deprecation_list_url = "{0}/{1}.yml/raw?ref=master".format(
            self.target_settings["iib_deprecation_list_url"].rstrip("/"), version.replace(".", "_")
        )
        registry_url = self.target_settings["docker_settings"]["docker_reference_registry"][0]

        LOG.info("Getting the deprecation list for OCP version {0}".format(version))
        session = _get_requests_session()
        response = session.get(url=deprecation_list_url)
        if not response.ok:
            LOG.error(
                "Could not retrieve deprecation list after multiple attempts."
                " Status Code {0}".format(response.status_code)
            )
            response.raise_for_status()

        try:
            yaml_response = yaml.safe_load(response.text)
            if yaml_response:
                deprecation_list = [
                    "{0}/{1}".format(registry_url, bundle_path)
                    for pkg_deprecation_list in yaml_response.values()
                    for bundle_path in pkg_deprecation_list
                ]
        except Exception:
            LOG.error("Data in {0} is invalid".format(deprecation_list_url))
            raise

        LOG.info("Deprecation list retrieved successfully")
        return sorted(deprecation_list)

    def iib_add_bundles(self, bundles, archs, ocp_version, deprecation_list=None):
        """
        Construct and execute pubtools-iib command to add bundles to index image.

        Args:
            bundles ([str]):
                External URLs to bundle images to be added to the index image.
            archs ([str]):
                Architectures to build for.
            ocp_version (str):
                OCP version to add the bundles to. It acts as a tag of the index image.
            deprecation_list ([str]):
                List of bundles to be deprecated.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to add bundles '{0}' to index image version '{1}'".format(
                bundles, ocp_version
            )
        )
        args = ["--skip-pulp"]

        args += ["--iib-server", self.target_settings["iib_server"]]
        args += ["--iib-krb-principal", self.target_settings["iib_krb_principal"]]

        if "iib_overwrite_from_index" in self.target_settings:
            args += ["--overwrite-from-index"]
        if "iib_krb_ktfile" in self.target_settings:
            args += ["--iib-krb-ktfile", self.target_settings["iib_krb_ktfile"]]

        index_image = "{image_repo}:{tag}".format(
            image_repo=self.target_settings["iib_index_image"], tag=ocp_version
        )
        args += ["--index-image", index_image]
        for bundle in bundles:
            args += ["--bundle", bundle]
        for arch in archs:
            args += ["--arch", arch]
        # inconsistent way of presenting multiple arguments...
        if deprecation_list:
            args += ["--deprecation-list", ",".join(deprecation_list)]

        env_vars = {}
        if "iib_overwrite_from_index_token" in self.target_settings:
            env_vars["OVERWRITE_FROM_INDEX_TOKEN"] = self.target_settings[
                "iib_overwrite_from_index_token"
            ]

        return run_entrypoint(
            ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
            "pubtools-iib-add-bundles",
            args,
            env_vars,
        )

    def run_tag_images(self, source_ref, dest_refs, all_arch):
        """
        Prepare the "tag images" entrypoint with all the necessary arguments and run it.

        Args:
            source_ref (str):
                Source image reference.
            dest_refs ([str]):
                List of destination references.
            all_arch (bool):
                Whether all architectures should be copied.
        """
        tag_images(
            source_ref,
            dest_refs,
            all_arch=all_arch,
            quay_user=self.target_settings["quay_user"],
            quay_password=self.target_settings["quay_password"],
            remote_exec=True,
            send_umb_msg=True,
            ssh_remote_host=self.target_settings["ssh_remote_host"],
            ssh_username=self.target_settings["ssh_user"],
            ssh_password=self.target_settings["ssh_password"],
            umb_urls=self.target_settings["docker_settings"]["umb_urls"],
            umb_cert=self.target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            # assumption that we'll continue using .pem format
            umb_client_key=self.target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            umb_ca_cert=self.target_settings["docker_settings"].get(
                "umb_ca_cert", "/etc/pki/tls/certs/ca-bundle.crt"
            ),
        )

    @log_step("Build index images")
    def build_index_images(self):
        """
        Perform the 'build' part of the operator workflow.

        The workflow can be summarized as:
        - Use Pyxis to parse 'com.redhat.openshift.versions'
        - Create mapping of which bundles should be pushed to which index image versions
        - Contact IIB to add the bundles to the index images

        Returns ({str:dict}):
            Dictionary containing IIB results and signing keys for all OPM versions. Data will be
            used in operator signing. Dictionary structure:
            {
                "version": {
                    "iib_result": (...) (object returned by iiblib)
                    "signing_keys": [...] (list of signing keys to be used for signing)
                }
            }

        """
        version_items_mapping = self.generate_version_items_mapping()
        iib_results = {}

        for version, items in sorted(version_items_mapping.items()):
            bundles = [self.public_bundle_ref(i) for i in items]
            all_archs = [
                i.metadata["arch"] if i.metadata["arch"] != "x86_64" else "amd64" for i in items
            ]
            archs = sorted(list(set(all_archs)))
            signing_keys = sorted(list(set([item.claims_signing_key for item in items])))

            # Get deprecation list
            deprecation_list = self.get_deprecation_list(version)

            # build index image in IIB
            build_details = self.iib_add_bundles(bundles, archs, version, deprecation_list)

            iib_results[version] = {"iib_result": build_details, "signing_keys": signing_keys}

        return iib_results

    @log_step("Push index images to Quay")
    def push_index_images(self, iib_results):
        """
        Push index images which were built in the previous stage to Quay.

        Args:
            iib_results (dict):
                IIB results returned by the build stage
        """
        image_schema = "{host}/{namespace}/{repo}"
        index_image_repo = image_schema.format(
            host=self.quay_host,
            namespace=self.target_settings["quay_namespace"],
            repo=get_internal_container_repo_name(self.target_settings["quay_operator_repository"]),
        )

        for version in sorted(iib_results.keys()):
            build_details = iib_results[version]["iib_result"]

            _, tag = build_details.index_image.split(":", 1)
            dest_image = "{0}:{1}".format(index_image_repo, tag)
            self.run_tag_images(build_details.index_image, [dest_image], True)



import logging

from .utils.misc import setup_arg_parser, add_args_env_variables
from .manifest_list_merger import ManifestListMerger

LOG = logging.getLogger("PubLogger")
logging.basicConfig()
LOG.setLevel(logging.INFO)

MERGE_MANIFEST_LIST_ARGS = {
    ("--source-ref",): {
        "help": "Source image reference. Manifest list data of this image will "
        "overwrite destination's manifest list",
        "required": True,
        "type": str,
    },
    ("--dest-ref",): {
        "help": "Destination image reference. Must be specified by tag. New manifest list will be "
        "uploaded to this image reference.",
        "required": True,
        "type": str,
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": True,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
}


def verify_merge_manifest_list_args(args):
    """Verify the presence and correctness of input parameters."""
    if "@" in args.dest_ref:
        raise ValueError("Destination must be specified via tag, not digest")

    if not args.quay_password:
        raise ValueError("Quay password must be set")


def merge_manifest_list_main(sysargs=None):
    """Entrypoint for manifest list merging."""
    parser = setup_arg_parser(MERGE_MANIFEST_LIST_ARGS)
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, MERGE_MANIFEST_LIST_ARGS)

    verify_merge_manifest_list_args(args)
    merger = ManifestListMerger(args.source_ref, args.dest_ref, args.quay_user, args.quay_password)
    merger.merge_manifest_lists()



from copy import deepcopy
import logging

from .quay_client import QuayClient

LOG = logging.getLogger("PubLogger")
logging.basicConfig()
LOG.setLevel(logging.INFO)


class ManifestListMerger:
    """Class containing logic for merging manifest lists of two images."""

    def __init__(self, src_image, dest_image, quay_username=None, quay_password=None, host=None):
        """
        Initialize.

        Args:
            src_image (str):
                Address to a new image whose manifest list contains the newer data.
            dest_image (str):
                Address to an older image whose data will be overwritten.
            quay_username (str):
                Username to login to quay. If ommited, external client instance should be set.
            quay_password (str):
                Password to login to quay. If ommited, external client instance should be set.
            host (str):
                Custom hostname to connect to. If ommited, standard quay.io will be used.
        """
        self.src_image = src_image
        self.dest_image = dest_image
        if quay_username and quay_password:
            self._quay_client = QuayClient(quay_username, quay_password, host)
        else:
            self._quay_client = None

    def set_quay_client(self, quay_client):
        """
        Set client instance to be used for the HTTP API operations.

        Args:
            quay_client (QuayClient):
                Instance of QuayClient.
        """
        self._quay_client = quay_client

    def merge_manifest_lists(self):
        """Merge manifest lists and upload to Quay. Main entrypoint method."""
        if not self._quay_client:
            raise RuntimeError("QuayClient instance must be set")

        LOG.info(
            "Merging manifest lists of images '{0}' and '{1}'".format(
                self.src_image, self.dest_image
            )
        )
        src_manifest_list = self._quay_client.get_manifest(self.src_image, manifest_list=True)
        dest_manifest_list = self._quay_client.get_manifest(self.dest_image, manifest_list=True)

        missing_archs = self.get_missing_architectures(src_manifest_list, dest_manifest_list)
        new_manifest_list = self._add_missing_architectures(src_manifest_list, missing_archs)

        LOG.info("Uploading the new manifest list to '{0}'".format(self.dest_image))
        self._quay_client.upload_manifest(new_manifest_list, self.dest_image)
        LOG.info("Merging manifests lists: complete.")

    @staticmethod
    def get_missing_architectures(src_manifest_list, dest_manifest_list):
        """
        Get architectures which are missing from the new source image.

        NOTE: this method assumes that images are built only for one OS. The following logic
        would need to be overwritten if multiple OS builds started to be made.

        Args:
            src_manifest_list (dict):
                Manifest list of the source image.
            dest_manifest_list (dict):
                Manifest list of the destination image.
        Returns ([dict]):
            List of arch manifest data present in destination image but missing from source.
        """
        missing_archs = []
        missing_archs_log = []
        src_archs = [arch["platform"]["architecture"] for arch in src_manifest_list["manifests"]]

        for dest_arch_dict in dest_manifest_list["manifests"]:
            if dest_arch_dict["platform"]["architecture"] not in src_archs:
                missing_archs.append(deepcopy(dest_arch_dict))
                missing_archs_log.append(dest_arch_dict["platform"]["architecture"])

        LOG.info(
            "Architectures missing from the new image: {0}".format(", ".join(missing_archs_log))
        )
        return missing_archs

    def _add_missing_architectures(self, src_manifest_list, missing_archs):
        """
        Add missing architectures to the source manifest list.

        Args:
            src_manifest_list (dict):
                Source manifest list.
            missing_archs ([dict]):
                Manifest data of missing architectures.
        Retuns (dict):
            New manifest list containing all the architectures.
        """
        new_manifest_list = deepcopy(src_manifest_list)
        new_manifest_list["manifests"] = new_manifest_list["manifests"] + missing_archs

        return new_manifest_list



import contextlib
import logging
import shlex
import subprocess
from six.moves import shlex_quote

import paramiko

LOG = logging.getLogger("PubLogger")
LOG.setLevel(logging.INFO)


# Python 2.6 version of paramiko doesn't support the usage
# of SSHClient as a context manager. This wrapper adds the functionality
@contextlib.contextmanager
def open_ssh_client():
    """Use SSHClient as a context manager."""
    client = paramiko.client.SSHClient()
    try:
        yield client
    finally:
        client.close()


class Executor(object):
    """
    Base executor class.

    Implementation of command execution should be done in
    descendant classes. Common pre- and post-processing should be
    implemented in this class.
    """

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """Run a bash command."""
        raise NotImplementedError  # pragma: no cover"

    def skopeo_login(self, username=None, password=None):
        """
        Attempt to login to Quay if no login credentials are present.

        Args:
            username (str):
                Username for login.
            password (str):
                Password for login.
        """
        cmd_check = "skopeo login --get-login quay.io"
        out, err = self._run_cmd(cmd_check, tolerate_err=True)
        if "not logged into" not in err and "not logged into" not in out:
            LOG.info("Already logged in to Quay.io")
            return

        if not username or not password:
            raise ValueError(
                "Skopeo login credentials are not present. Quay user and token must be provided."
            )
        LOG.info("Logging in to Quay with provided credentials")

        cmd_login = (
            "skopeo login --authfile $HOME/.docker/config.json -u {0} --password-stdin quay.io"
        ).format(shlex_quote(username))
        out, err = self._run_cmd(cmd_login, stdin=password)

        if "Login Succeeded" in out:
            LOG.info("Login successful")
        else:
            raise RuntimeError(
                "Login command didn't generate expected output. "
                "STDOUT: '{0}', STDERR: '{1}'".format(out, err)
            )

    def tag_images(self, source_ref, dest_refs, all_arch=False):
        """
        Copy image from source to destination(s) using skopeo.

        Args:
            source_ref (str):
                Reference of the source image.
            dest_refs ([str]):
                List of target references to copy the image to.
            all_arch (bool):
                Whether to copy all architectures (if multiarch image)
        """
        if all_arch:
            cmd = "skopeo copy --all docker://{0} docker://{1}"
        else:
            cmd = "skopeo copy docker://{0} docker://{1}"

        for dest_ref in dest_refs:
            LOG.info("Tagging source '{0}' to destination '{1}'".format(source_ref, dest_ref))
            self._run_cmd(cmd.format(shlex_quote(source_ref), shlex_quote(dest_ref)))
            LOG.info("Destination image {0} has been tagged.".format(dest_ref))

        LOG.info("Tagging complete.")


class LocalExecutor(Executor):
    """Run commands locally."""

    def __init__(self, params={}):
        """
        Initialize.

        Args:
            params (dict):
                Custom parameters to be applied when running the shell commands.
        """
        self.params = params
        self.params.setdefault("universal_newlines", True)
        self.params.setdefault("stderr", subprocess.PIPE)
        self.params.setdefault("stdout", subprocess.PIPE)
        self.params.setdefault("stdin", subprocess.PIPE)

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """
        Run a command locally.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (str):
                String to send to standard input for a command.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."

        p = subprocess.Popen(shlex.split(cmd), **self.params)
        out, err = p.communicate(input=stdin)

        if p.returncode != 0 and not tolerate_err:
            LOG.error("Command {0} failed with {1}".format(cmd, err))
            raise RuntimeError(err_msg)

        return out, err


class RemoteExecutor(Executor):
    """Run commands remotely via SSH."""

    def __init__(
        self,
        hostname,
        username=None,
        key_filename=None,
        password=None,
        port=None,
        accept_unknown_host=True,
    ):
        """
        Initialize.

        Args:
            hostname (str):
                Host to connect to.
            username (str):
                Username to authenticate as. Defaults to local username.
            key_filename (str):
                Path to a private key for authentication. Default location will be used if omitted.
            password (str):
                Password for ssh authentication. Has lower precedence than private key.
            port (int):
                Optional port of the host.
            accept_unknown_host (bool):
                Whether to accept an unknown host key. Defaults to True.
        """
        self.hostname = hostname
        self.username = username
        self.key_filename = key_filename
        self.password = password
        if accept_unknown_host:
            self.missing_host_policy = paramiko.client.WarningPolicy()
        else:
            self.missing_host_policy = paramiko.client.RejectPolicy()
        self.port = port if port else 22

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """
        Run a command remotely via SSH.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (str):
                String to send to standard input for a command.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."
        with open_ssh_client() as client:
            client.load_system_host_keys()
            client.set_missing_host_key_policy(self.missing_host_policy)
            client.connect(
                self.hostname,
                username=self.username,
                port=self.port,
                password=self.password,
                key_filename=self.key_filename,
            )

            ssh_in, out, err = client.exec_command(cmd)
            if stdin:
                ssh_in.channel.send(stdin)
                ssh_in.channel.shutdown_write()

            out_text = out.read().decode("utf-8")
            err_text = err.read().decode("utf-8")
            if out.channel.recv_exit_status() != 0 and not tolerate_err:
                LOG.error("Command {0} failed with {1}".format(cmd, err_text))
                raise RuntimeError(err_msg)

        return out_text, err_text
